#include <wayland-server.h>
#include <glm/glm.hpp>

#include <string>
#include <vector>
#include <utility>
#include <memory>
#include <functional>
#include <algorithm>
#include <sstream>
#include <map>
#include <cmath>
#include <cstddef>
